import logging
from functools import wraps as _wraps
from functools import partial as _partial

from typing import Optional as _Optional
from typing import List as _List
from collections.abc import Callable as _Callable
from collections import namedtuple

import attr

from stag.ecs import Page, Site


log = logging.getLogger(__name__)


def run_always(*a, **kw):
    """Helper function which can be used when registering plugins to mark that
    they should be run for all files."""
    return True


_Cond = _Callable[[Page], bool]
_Plugin = _Callable
_PlugConfig = _Optional[object]


@attr.s(auto_attribs=True)
class PlugData:
    name: str
    plugin: _Plugin
    condition: _Cond
    config: _PlugConfig


_READERS = {}
_GENERATORS = {}
_WRITERS = {}


def _register_impl(out: list, name: str, *a) -> None:
    if name in out:
        raise KeyError(f"name conflict: {name}")
    out[name] = PlugData(name, *a)


def _register(
    where: dict, name: str, cond: _Cond = run_always, config: _PlugConfig = None
):
    def decor(fn):
        _register_impl(where, name, fn, cond, config)

        @_wraps(fn)
        def wrapper(*a, **kw):
            return fn(*a, **kw)

        return wrapper

    return decor


def _get_enabled_in_order(plugins, order=None):
    if order is None:
        return list(plugins.values())

    ret = []
    for name in order:
        try:
            ret.append(plugins[name])
        except KeyError:
            log.error(f"Plugin not found: {name}")
            continue
    return ret


reader = _partial(_register, _READERS)
generator = _partial(_register, _GENERATORS)
writer = _partial(_register, _WRITERS)

readers = _partial(_get_enabled_in_order, _READERS)
generators = _partial(_get_enabled_in_order, _GENERATORS)
writers = _partial(_get_enabled_in_order, _WRITERS)



from stag.ecs import Path, Page, Metadata


_NO_TYPE = "_NO_TYPE"


def make_meta(*types, content=""):
    for t in types:
        if t is _NO_TYPE:
            yield Metadata()
        else:
            yield Metadata(type=t)


def make_pages(*metadatas):
    # not set to ease assertions
    return [Page(url=f"/{i}", metadata=md) for i, md in enumerate(metadatas)]


def test_page_filter_any(site):
    site.pages = make_pages(*make_meta(None, "page", "page", "post", _NO_TYPE), None)
    filtered = list(site.filter_pages())
    assert filtered == site.pages[0:5]


def test_page_filter_none(site):
    site.pages = make_pages(*make_meta(None, "page", "page", "post", _NO_TYPE), None)
    filtered = list(site.filter_pages(None))
    assert filtered == [site.pages[0]]


def test_page_filter_type(site):
    site.pages = make_pages(*make_meta(None, "page", "page", "post", _NO_TYPE), None)
    filtered = list(site.filter_pages("page"))
    assert filtered == [site.pages[1], site.pages[2]]


def test_page_filter_unexisting_type(site):
    site.pages = make_pages(*make_meta(None, "page", "page", "post", _NO_TYPE), None)
    filtered = list(site.filter_pages("someethingelse"))
    assert filtered == []



from functools import cache

from typing import Optional as _Optional

import attr
from jinja2 import Template, Environment, ChoiceLoader, FileSystemLoader
from jinja2.exceptions import TemplateSyntaxError, TemplateNotFound

from .base import generator


class MacroError(Exception):
    pass


def raise_macro_error(msg):
    raise MacroError(msg)


@attr.s(auto_attribs=True)
class MacrosConfig:
    path: _Optional[str] = None


def cond(page):
    return page.source and page.input and page.input.content


@cache
def get_env(macros_dir, theme_dir):
    loaders = [FileSystemLoader(dir) for dir in [macros_dir, theme_dir] if dir]
    env = Environment(
        loader=ChoiceLoader(loaders), trim_blocks=True, lstrip_blocks=True
    )
    env.globals["raise"] = raise_macro_error

    return env


@generator("macros", cond=cond, config=MacrosConfig())
def resolve_macros(page, site):
    config = site.config
    myconfig = config.plugins.macros

    env = get_env(myconfig.path, config.plugins.theme.name)
    env.globals["site"] = site
    env.globals["print"] = print

    try:
        template = env.from_string(page.input.content)
        page.input.content = template.render(page=page, site=site)
    except TemplateNotFound as e:
        raise MacroError(f"{page.source.relpath}: missing macro definitions file: {e}")
    except Exception as e:
        raise MacroError(f"{page.source.relpath}: {e}")



import os
import sys
import argparse
import shutil
import fnmatch
import logging
import pkgutil
from copy import deepcopy
from itertools import chain
from datetime import datetime

import tomli
from slugify import slugify

from stag.config import Config, TaxonomyTable
from stag.ecs import Page, Site, Path, Taxonomy, Metadata, Content
from stag.plugins.base import readers, generators, writers
from stag.utils import chdir, SortDate
from stag._version import version

log = logging.getLogger(__name__)


def apply_plugins(plugins, site):
    for page in site.pages:
        for plug in plugins:
            if plug.condition(page):
                log.debug(f"Running {plug.name} on page {page.url}")
                plug.plugin(page, site)


def load_plugins(path):
    for finder, name, ispkg in pkgutil.iter_modules([path]):
        mod = finder.find_module(name)
        if mod:
            mod.load_module()


def copy_output_type(pages, default="html"):
    for page in pages:
        if page.output and page.output.type:
            return page.output.type
    return default


def update_metadata(page, **kw):
    if not page.metadata:
        page.metadata = Metadata()
    kw.update(page.metadata.data)
    page.metadata.data = kw


def make_taxonomies(site):
    if not site.config.taxonomies:
        return

    taxonomies = {
        ttab.key: Taxonomy(key=ttab.key, singular=ttab.singular, plural=ttab.plural)
        for ttab in site.config.taxonomies
    }

    # We split the algorithm of creating taxonomies to 2 parts:
    #   1. gathering all the taxonomies, together with creating appropriate
    #      components
    #   2. creating page, which can be optionally already created (see below)
    #
    # This could be technically done in only one pass, but would require
    # Page.list to become a dictionary to nod add a complexity of searching
    # of term urls through the list/set of Pages

    for page in site.pages:
        if not page.metadata:
            continue
        for taxonomy in taxonomies:
            terms = page.metadata.get(taxonomy)
            if not terms:
                continue
            if not isinstance(terms, list):
                terms = [terms]
            for term in terms:
                taxonomies[taxonomy][term].add(page)

    # Pages can be created by users for taxonomies and terms. In this situation
    # stag will override some members (page.list), update others with unset
    # values (metadata) and leave the rest untouched (e.g. content).
    #
    # This is to give users possibility to hardcode custom metadata for specific
    # taxonomies.
    now = datetime.now()
    for base, taxonomy in taxonomies.items():
        baseslug = slugify(base)
        taxo_page = site.get_or_create_page(baseslug)
        taxo_page.list = []
        taxo_page.taxonomy = taxonomy
        update_metadata(taxo_page, title=taxonomy.plural, date=now)

        for term, pages in taxonomy.terms.items():
            termslug = slugify(term)
            term_page = site.get_or_create_page(f"{baseslug}/{termslug}")
            term_page.list = sorted(pages, key=SortDate)
            term_page.output = Content(type=copy_output_type(pages))
            update_metadata(term_page, title=term, date=now, taxonomy=base)

            taxo_page.list.append(term_page)

        taxo_page.list.sort()
        taxo_page.output = Content(type=copy_output_type(taxo_page.list))

    site.taxonomies = list(taxonomies.values())


def build(args, config):
    log.info(f"Building site to {config.output}")
    site = Site(config=config)

    rds = readers(config.readers)
    roots = [
        config.content,
        os.path.join(config.plugins.theme.name, "static"),
        "static",
    ]

    for root in roots:
        gather_files(root, rds, site)

    shutil.rmtree(config.output, ignore_errors=True)
    os.makedirs(config.output)

    apply_plugins(generators(config.generators), site)

    make_taxonomies(site)

    apply_plugins(writers(config.writers), site)


def serve(args, config):
    from http.server import HTTPServer, SimpleHTTPRequestHandler

    build(args, config)
    with chdir(config.output):
        log.info(f"Running simple HTTP server on http://localhost:{args.port}.")
        server_address = ("", args.port)
        httpd = HTTPServer(server_address, SimpleHTTPRequestHandler)
        httpd.serve_forever()


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    else:
        slug, _ = os.path.splitext(path.basename)
        return os.path.join(path.reldirname, pat.filebase)


def gather_files(srcdir, plugins, site):
    def read_file(path):
        for plug in plugins:
            if plug.condition(path):
                log.debug(f"Reading {path.path} with {plug.name} plugin.")
                return plug.plugin(path, site)
        return None

    for curdir, _, files in os.walk(srcdir):
        for f in files:
            path = Path(os.path.join(curdir, f), srcdir)
            page = read_file(path)
            if page:
                site.add_page(page)
            else:
                log.warning(f"No valid reader for {path.path}")


def read_config(path):
    try:
        with open(path) as f:
            config = tomli.loads(f.read())
    except FileNotFoundError:
        config = {}

    return config


def make_root_config(cfg_dct):
    recognised = {}
    supported = Config.__attrs_attrs__  # bare tuple with properties for names

    user_dct = cfg_dct.pop("user", {})
    taxo_lst = cfg_dct.pop("taxonomies", [])

    for key, val in cfg_dct.items():
        if not hasattr(supported, key):
            log.error(
                f"Unrecognised option: {key}. User-defined options should be put in the [user] table."
            )
            continue
        recognised[key] = val

    cfg = Config(**recognised)
    cfg.user.__dict__.update(user_dct)
    cfg.taxonomies.extend(TaxonomyTable(**t) for t in taxo_lst)
    return cfg


def update_config_plugins(cfg, plugins_dct):
    for plugin in chain(readers(), generators(), writers()):
        if plugin.config:
            pconf = deepcopy(plugin.config)
            for key, val in plugins_dct.get(plugin.name, {}).items():
                if not hasattr(pconf, key):
                    log.error(f"Unrecognised option {key} for {plugin.name} plugin")
                    continue

                if isinstance(val, dict):
                    getattr(pconf, key).update(val)
                else:
                    setattr(pconf, key, val)
            setattr(cfg.plugins, plugin.name, pconf)

    return cfg


def prepare_args():
    parser = argparse.ArgumentParser(description="Simply Stupid Static Site Generator")
    parser.set_defaults(verbosity=logging.INFO)

    parser.add_argument(
        "-c",
        "--config",
        nargs="?",
        default="config.toml",
        help="path to stag's configuration file",
    )

    parser.add_argument(
        "-D",
        "--debug",
        action="store_const",
        const=logging.DEBUG,
        dest="verbosity",
        help="show debug messages",
    )

    parser.add_argument("--version", action="version", version=f"%(prog)s {version}")

    sp = parser.add_subparsers(required=True, dest="subcommands")

    sp_build = sp.add_parser("build")
    sp_build.set_defaults(func=build)

    sp_serve = sp.add_parser("serve")
    sp_serve.add_argument("-p", "--port", type=int, default="8000", help="HTTP port")
    sp_serve.set_defaults(func=serve)

    return parser.parse_args()


def main():
    args = prepare_args()
    logging.basicConfig(format="%(message)s", level=args.verbosity)

    cfg_dct = read_config(args.config)
    plugins_dct = cfg_dct.pop("plugins", None)

    cfg = make_root_config(cfg_dct)
    load_plugins(cfg.pluginspath)
    update_config_plugins(cfg, plugins_dct)

    try:
        return args.func(args, cfg)
    except Exception as e:
        log.error(f"Critical error: {e}")
        if args.verbosity == logging.DEBUG:
            import pdb, traceback

            extype, value, tb = sys.exc_info()
            traceback.print_exc()
            pdb.post_mortem(tb)
            raise
        return 1



import os
import logging
from functools import cache

import attr
from jinja2 import Template, Environment, FileSystemLoader, FunctionLoader, ChoiceLoader

from .base import writer
from ._jinja_functions import update_env

try:
    from stag._version import version
except ImportError:
    version = "unknown version"


log = logging.getLogger(__name__)


_DEFAULT_TEMPLATE = "{{ content }}"
_DEFAULT_HTML_TEMPLATE = f"""<!DOCTYPE html>
<html lang="{{{{ site.config.get('lang', 'en') }}}}">
  <head>
    <meta name="generator" content="stag ({version})" />
    <meta charset="UTF-8">
    <title>{{{{ page.metadata.title }}}} - {{{{ site.config.title }}}}</title>
  </head>
  <body>
  {{{{ content }}}}
  </body>
</html>
"""


def load_default_template(name):
    log.error(f"Template not found: {name}. Using built-in basic template.")
    _, ext = os.path.splitext(name)
    template = _DEFAULT_HTML_TEMPLATE if ext == ".html" else _DEFAULT_TEMPLATE
    return template


@attr.s(auto_attribs=True)
class JinjaConfig:
    name: str = os.path.join("themes", "default")
    templates: dict = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        default_templates = dict(page="page", list="list", taxonomy="taxonomy")
        default_templates.update(self.templates)
        self.templates = default_templates


def outputable(page):
    return page.output and page.metadata


@cache
def get_env(theme):
    env = Environment(
        loader=ChoiceLoader(
            [
                FileSystemLoader(theme),
                FunctionLoader(load_default_template),
            ]
        ),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    update_env(env)
    return env


def is_list(page):
    return bool(page.list)


def is_taxonomy(page):
    return bool(page.taxonomy)


def get_default_type(page, templates):
    if is_taxonomy(page):
        return templates["taxonomy"]
    if is_list(page):
        return templates["list"]
    return templates["page"]


def render_page(page, site, env):
    config = site.config
    myconfig = config.plugins.theme

    type_ = page.metadata.get("type")
    if not type_:
        type_ = get_default_type(page, myconfig.templates)

    template = env.get_template(f"{type_}.{page.output.type}")
    url = page.url.strip("/")
    filename = f"index.{page.output.type}"

    outpath = os.path.join(config.output, url, filename)
    outdir = os.path.dirname(outpath)

    if os.path.exists(outpath):
        log.error(f"page already exists, skipping: {outpath}")
        return

    os.makedirs(outdir, exist_ok=True)
    with open(outpath, "w") as fd:
        fd.write(template.render(content=page.output.content, site=site, page=page))


@writer("theme", cond=outputable, config=JinjaConfig())
def render(page, site):
    env = get_env(site.config.plugins.theme.name)
    return render_page(page, site, env)



import os

import attr
import tomli
import markdown

from .base import reader, generator
from stag.ecs import Page, Content, Metadata


def is_md(path):
    return path.ext == "md"


def is_opened_md(page):
    return page.input and page.input.type == "md"


@attr.s(auto_attribs=True)
class MarkdownConfig:
    extensions: list[str] = ["footnotes", "fenced_code", "smarty"]


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    return os.path.join(path.reldirname, path.filebase)


@reader("markdown", cond=is_md)
def read(path, site):
    frontmatter = "+++"

    frontmatter_parsed = False
    in_frontmatter = False
    metadata = []
    content = []

    with open(path.path) as fd:
        for line in fd:
            line = line.rstrip("\r\n")
            if line == frontmatter:
                if not frontmatter_parsed:
                    frontmatter_parsed = True
                    in_frontmatter = True
                    continue
                if in_frontmatter:
                    in_frontmatter = False
                    continue

            if in_frontmatter:
                metadata.append(line)
            else:
                content.append(line)

    metadata_parsed = tomli.loads("\n".join(metadata))
    content_parsed = "\n".join(content)

    url = deduce_url(path)
    metadata = Metadata(metadata_parsed)
    input_ = Content("md", content_parsed)
    return Page(url=url, source=path, metadata=metadata, input=input_)


@generator("markdown", cond=is_opened_md, config=MarkdownConfig())
def generate(page, site):
    myconfig = site.config.plugins.markdown

    assert "title" in page.metadata, f"No title in {page.source.relpath}"

    extensions = myconfig.extensions
    html = markdown.markdown(page.input.content, extensions=extensions)
    page.output = Content("html", html)



import os
from contextlib import contextmanager as _contextmanager
from datetime import date as _date_t
from datetime import datetime as _datetime_t


@_contextmanager
def chdir(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


class SortDate:
    def __init__(self, page):
        self.ts = None
        if page.metadata:
            self.ts = page.metadata.timestamp

    def __lt__(self, other):
        if self.ts and other.ts:
            return self.ts < other.ts
        return self.ts and not other.ts



import os
import shutil

from .base import reader, writer
from stag.ecs import Page


class CopyError(Exception):
    pass


def write_cnd(page):
    return page.source and not page.input


@reader("static")
def read(path, site):
    def strip_roots(url):
        opj = os.path.join
        c = site.config
        excls = ["static", opj(c.plugins.theme.name, "static"), c.content]
        for ex in excls:
            if url.startswith(ex):
                return url[len(ex) :].strip("/")
        return url.strip("/")

    url = path.relpath
    return Page(url=f"/{url}", source=path)


@writer("copy", cond=write_cnd)
def write(page, site):
    url = page.url.strip("/")
    output_path = os.path.join(site.config.output, url)

    if os.path.exists(output_path):
        raise CopyError(f"File exists: {output_path}")

    target_dir = os.path.dirname(output_path)
    os.makedirs(target_dir, exist_ok=True)
    shutil.copy(page.source.path, target_dir)



from __future__ import annotations

import os
from collections import UserDict, defaultdict
from datetime import date as _date_t
from datetime import datetime as _datetime_t

from typing import Union as _Union
from typing import Optional as _Optional
from typing import Any as _Any
from collections.abc import Mapping as _Mapping

from stag.config import TaxonomyTable as _TaxonomyTable

import attr
from dateutil.parser import parse as _parse_dt


# sentinel used for parameters for which None is a valid value
_ANY = "_ANY"


def _urlize(url):
    url = url.strip("/")
    return f"/{url}"


@attr.s(auto_attribs=True, frozen=True)
class Path:
    path: str
    root_dir: str

    @property
    def relpath(self):
        return self.path[len(self.root_dir) :].strip("/")

    @property
    def ext(self):
        return os.path.splitext(self.path)[1][len(os.extsep) :]

    @property
    def filebase(self):
        return os.path.splitext(self.basename)[0]

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def reldirname(self):
        return os.path.dirname(self.relpath)


@attr.s(auto_attribs=True)
class Content:
    type: str
    content: _Any = attr.ib(None, repr=False)


class Metadata(UserDict):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._normalize_data()

    def _normalize_data(self):
        title = self.data.get("title", "")
        date = self.data.get("date")
        if date:
            if isinstance(date, (int, str, float)):
                self.data["date"] = _parse_dt(date)
            elif type(date) == _date_t:
                self.data["date"] = _datetime_t.fromordinal(date.toordinal())
            self.data["timestamp"] = self.data["date"].timestamp()

        self.__dict__.update(self.data)


def _terms_factory():
    return defaultdict(set)


@attr.s(auto_attribs=True)
class Taxonomy:
    key: str
    singular: str
    plural: str
    terms: _Mapping[str, set[Page]] = attr.ib(factory=_terms_factory, repr=False)

    def __getitem__(self, key):
        return self.terms[key]

    def __setitem__(self, key, val):
        self.terms[key] = val


@attr.s(auto_attribs=True, cmp=False, hash=False)
class Page:
    _url: str = attr.ib(converter=_urlize)
    metadata: _Optional[Metadata] = attr.ib(default=None, repr=False)
    source: _Optional[Path] = attr.ib(default=None, repr=False)
    input: _Optional[Content] = attr.ib(default=None, repr=False)
    output: _Optional[Content] = attr.ib(default=None, repr=False)
    taxonomy: _Optional[Taxonomy] = attr.ib(default=None, repr=False)
    list: _Optional[list[Page]] = attr.ib(default=None, repr=False)

    @property
    def url(self):
        return self._url

    def __hash__(self):
        return hash(self._url)

    def __eq__(self, other):
        return self._url == other._url

    def __lt__(self, other):
        return self._url < other._url


@attr.s(auto_attribs=True)
class Site:
    config: dict = attr.ib(repr=False)
    _pages: _Mapping[str, Page] = attr.ib(factory=dict, repr=False)
    taxonomies: _Optional[list[Taxonomy]] = attr.ib(None, repr=False)

    @property
    def pages(self):
        return list(self._pages.values())

    @property
    def ordinary_pages(self):
        for page in self._pages.values():
            if page.source and page.input and page.output and page.metadata:
                yield page

    def subpages_of(self, val):
        base = _urlize(val)
        for page in self.ordinary_pages:
            cmp, _, _ = page.url.rpartition("/")
            if cmp == base:
                yield page

    def add_page(self, page):
        if page.url in self._pages:
            raise ValueError(f"URL {page.url} for file {path} not unique")
        self._pages[page.url] = page

    def get_or_create_page(self, url):
        return self._pages.setdefault(_urlize(url), Page(url=url))

    def filter_pages(self, ptype=_ANY):
        for page in self.pages:
            if page.metadata is not None and ptype is _ANY:
                yield page
            elif page.metadata is not None and page.metadata.get("type", _ANY) == ptype:
                yield page

    def find(self, url):
        return self._pages.get(_urlize(url))



from typing import Optional as _Optional

import attr


class PluginTable:
    def __str__(self):
        opts = ", ".join(k for k in self.__dict__)
        return f"PluginTable({opts})"

    def __repr__(self):
        opts = ", ".join(k for k in self.__dict__)
        return f"PluginTable({opts})"


class UserOptionsTable:
    def __str__(self):
        opts = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"UserOptionsTable({opts})"

    def __repr__(self):
        opts = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"UserOptionsTable({opts})"


@attr.s(auto_attribs=True)
class TaxonomyTable:
    key: str
    singular: _Optional[str] = None
    plural: _Optional[str] = None

    def __attrs_post_init__(self):
        if self.singular is None:
            self.singular = self.key
        if self.plural is None:
            self.plural = self.key


@attr.s(auto_attribs=True)
class Config:
    title: str = "MySite"
    url: str = "https://example.com"
    language: str = "en"
    pluginspath: str = "plugins"
    content: str = "content"
    output: str = "_output"
    readers: _Optional[list[str]] = ["markdown", "static"]
    generators: _Optional[list[str]] = ["macros", "markdown"]
    writers: _Optional[list[str]] = ["theme", "copy"]
    taxonomies: list[TaxonomyTable] = attr.ib(factory=list, init=False, repr=False)
    user: UserOptionsTable = attr.ib(factory=UserOptionsTable, repr=False)
    plugins: PluginTable = attr.ib(factory=PluginTable, init=False)

    def get(self, name, default=None):
        return getattr(self, name, default)
